@@ -9,8 +9,8 @@ from typing import (
     Generator,
     List,
     Optional,
-    cast,
     Iterable,
+    Iterator,
     Tuple,
     Type,
     Union,
@@ -20,13 +20,14 @@ from jira import JIRA, Issue
 from rich.progress import Progress, TaskID, BarColumn, TimeRemainingColumn
 
 from .cache import MinimumRecencyCache
-from .plugin import get_installed_functions
+from .plugin import BaseSource, get_installed_functions, get_installed_sources
 from .types import (
     ExpressionList,
     JqlList,
     QueryDefinition,
     SelectFieldDefinition,
     Expression,
+    WhereParamDict,
 )
 from .utils import (
     calculate_result_hash,
@@ -85,8 +86,9 @@ class Result(metaclass=ABCMeta):
     def evaluate_expression(
         self,
         expression: Expression,
-        group_by: ExpressionList = None,
-        functions: Dict[str, Callable] = None,
+        group_by: Optional[ExpressionList] = None,
+        field_name_map: Optional[Dict[str, str]] = None,
+        functions: Optional[Dict[str, Callable]] = None,
     ):
         if group_by is None:
             group_by = []
@@ -98,7 +100,9 @@ class Result(metaclass=ABCMeta):
         if expression_includes_group_by(expression, group_by):
             row_source = self.single()
 
-        return get_field_data(row_source, expression, functions)
+        return get_field_data(
+            row_source, expression, functions=functions, interpolations=field_name_map,
+        )
 
 
 class SingleResult(Result):
@@ -112,6 +116,18 @@ class SingleResult(Result):
         return self
 
 
+class GroupedFieldContainer(list):
+    def __getattr__(self, name):
+        results = GroupedFieldContainer()
+
+        for row in self:
+            value = getattr(row, name, None)
+            if value is not None:
+                results.append(value)
+
+        return results
+
+
 class GroupedResult(Result):
     def __init__(
         self, rows: List[SingleResult] = None,
@@ -145,11 +161,13 @@ class GroupedResult(Result):
             # Exclude empty rows -- in SQL when you run an aggregation
             # function on a set of rows, NULL rows are skipped, so we
             # should probably do the same?
-            result[field] = [
-                row.as_dict()[field]
-                for row in self._rows
-                if row.as_dict().get(field) is not None
-            ]
+            result[field] = GroupedFieldContainer(
+                [
+                    row.as_dict()[field]
+                    for row in self._rows
+                    if row.as_dict().get(field) is not None
+                ]
+            )
 
         return result
 
@@ -162,14 +180,15 @@ class Query:
     def _ensure_str(self, iterable=Iterable[Any]) -> List[str]:
         return [str(item) for item in iterable]
 
-    def _get_all_fields(self):
-        fields = [{"id": "key"}, {"id": "id"}] + self._jira.fields()
-        for field in fields:
-            definition: SelectFieldDefinition = {
-                "expression": field["id"],
-                "column": field["id"],
-            }
-            yield definition
+    def _get_all_fields(self) -> List[SelectFieldDefinition]:
+        sources = get_installed_sources()
+
+        try:
+            source = sources[self.from_]
+        except KeyError:
+            return []
+
+        return source.get_all_fields(self._jira)
 
     @property
     def select(self) -> List[SelectFieldDefinition]:
@@ -188,8 +207,11 @@ class Query:
         return self._definition["from"]
 
     @property
-    def where(self) -> JqlList:
-        return self._ensure_str(self._definition.get("where", []))
+    def where(self) -> Union[JqlList, WhereParamDict]:
+        where = self._definition.get("where", [])
+        if isinstance(where, list):
+            return self._ensure_str(where)
+        return where
 
     @property
     def order_by(self) -> JqlList:
@@ -294,6 +316,11 @@ class CounterChannel:
         return self._counter
 
 
+class FieldNameMap(dict):
+    def __missing__(self, key):
+        return key
+
+
 class Executor:
     def __init__(
         self,
@@ -312,6 +339,8 @@ class Executor:
         if enable_cache:
             self._cache = MinimumRecencyCache(get_cache_path())
 
+        self._field_name_map: Dict[str, str] = FieldNameMap()
+
     @property
     def jira(self) -> JIRA:
         return self._jira
@@ -328,6 +357,14 @@ class Executor:
     def functions(self) -> Dict[str, Callable]:
         return self._functions
 
+    @property
+    def field_name_map(self) -> Dict[str, str]:
+        if not self._field_name_map:
+            for jira_field in self.jira.fields():
+                self._field_name_map[jira_field["name"]] = jira_field["id"]
+
+        return self._field_name_map
+
     def _get_jql(self) -> str:
         query = " AND ".join(f"({q})" for q in self.query.where)
         order_by_fields = ", ".join(self.query.order_by)
@@ -337,15 +374,17 @@ class Executor:
 
         return query
 
-    def _get_issues(
-        self, task: TaskID, out_channel: CounterChannel
-    ) -> Generator[Result, None, None]:
-        jql = self._get_jql()
-        cache_key = f"{jql}:{','.join(self.query.order_by)}:{self.query.limit}"
-
-        start_at = 0
-        max_results = 2 ** 32
-        result_limit = self.query.limit or 2 ** 32
+    def _get_cached(self, source: BaseSource) -> Iterator[Result]:
+        cache_key = ":".join(
+            [
+                str(self.jira.client_info()),
+                str(self.query.from_),
+                str(self.query.where),
+                str(self.query.order_by),
+                str(self.query.limit),
+                str(self.query.expand),
+            ]
+        )
 
         if self.query.cache:
             try:
@@ -356,8 +395,8 @@ class Executor:
                 if not cached_results:
                     raise KeyError(cache_key)
 
-                out_channel.set(len(cached_results))
-                self.progress.remove_task(task)
+                source.update_count(len(cached_results))
+                source.remove_progress()
                 for result in cached_results:
                     yield SingleResult(Issue({}, None, result))
                 return
@@ -366,38 +405,9 @@ class Executor:
 
         cache = []
 
-        self.progress.update(task, completed=0, total=1, visible=True)
-        while start_at < min(max_results, result_limit):
-            results = self.jira.search_issues(
-                jql,
-                startAt=start_at,
-                expand=",".join(self.query.expand),
-                fields="*all",
-                maxResults=min(result_limit, 100),
-            )
-
-            max_results = results.total
-            count = min([results.total, result_limit])
-            out_channel.set(count)
-
-            for result in results:
-                result = cast(Issue, result)
-                self.progress.update(task, advance=1, total=count, visible=True)
-
-                if self.query.cache:
-                    cache.append(result.raw)
-
-                # Do not return the issue directly -- it'll have session
-                # information available that might allow you to accidentally
-                # do something that modifies data.  Instead, take the Raw
-                # JSON and use this as if we were re-hydrating from cache.
-                yield SingleResult(Issue({}, None, result.raw))
-
-                start_at += 1
-
-                # Return early if our result limit has been reached
-                if start_at >= result_limit:
-                    break
+        for result in source:
+            cache.append(result)
+            yield SingleResult(source.rehydrate(result))
 
         if self.query.cache:
             _, max_store = self.query.cache
@@ -406,11 +416,11 @@ class Executor:
 
     def _process_filter(
         self,
-        iterator: Generator[Result, None, None],
+        iterator: Iterator[Result],
         task: TaskID,
         input_channel: CounterChannel,
         output_channel: CounterChannel,
-    ) -> Generator[Result, None, None]:
+    ) -> Iterator[Result]:
         output_channel.zero()
 
         for row in iterator:
@@ -418,9 +428,7 @@ class Executor:
 
             include_row = True
             for filter_expression in self.query.filter:
-                if not row.evaluate_expression(
-                    filter_expression, self.query.group_by, self.functions,
-                ):
+                if not self.evaluate_expression(row, filter_expression):
                     include_row = False
                     break
 
@@ -432,11 +440,11 @@ class Executor:
 
     def _process_group_by(
         self,
-        iterator: Generator[Result, None, None],
+        iterator: Iterator[Result],
         task: TaskID,
         input_channel: CounterChannel,
         output_channel: CounterChannel,
-    ) -> Generator[Result, None, None]:
+    ) -> Iterator[Result]:
         groups: Dict[int, Result] = {}
 
         output_channel.zero()
@@ -458,11 +466,11 @@ class Executor:
 
     def _process_having(
         self,
-        iterator: Generator[Result, None, None],
+        iterator: Iterator[Result],
         task: TaskID,
         input_channel: CounterChannel,
         output_channel: CounterChannel,
-    ) -> Generator[Result, None, None]:
+    ) -> Iterator[Result]:
         output_channel.zero()
 
         for row in iterator:
@@ -470,9 +478,7 @@ class Executor:
 
             include_row = True
             for having in self.query.having:
-                if not row.evaluate_expression(
-                    having, self.query.group_by, self.functions,
-                ):
+                if not self.evaluate_expression(row, having):
                     include_row = False
                     break
 
@@ -484,11 +490,11 @@ class Executor:
 
     def _process_sort_by(
         self,
-        iterator: Generator[Result, None, None],
+        iterator: Iterator[Result],
         task: TaskID,
         input_channel: CounterChannel,
         output_channel: CounterChannel,
-    ) -> Generator[Result, None, None]:
+    ) -> Iterator[Result]:
         # First, materialize our list
         rows = list(iterator)
         output_channel.set(len(rows))
@@ -498,9 +504,7 @@ class Executor:
         for sort_expression, reverse in reversed(self.query.sort_by):
 
             def sort_key(row):
-                result = row.evaluate_expression(
-                    sort_expression, self.query.group_by, self.functions
-                )
+                result = self.evaluate_expression(row, sort_expression)
                 self.progress.update(task, advance=1)
 
                 return NullAcceptableSort(result)
@@ -513,16 +517,18 @@ class Executor:
         result: Dict[str, Any] = {}
 
         for field_defn in self.query.select:
-            result[field_defn["column"]] = row.evaluate_expression(
-                field_defn["expression"], self.query.group_by, self.functions,
+            result[field_defn["column"]] = self.evaluate_expression(
+                row, field_defn["expression"],
             )
 
         return result
 
     def _get_iterator(self) -> Generator[Dict[str, Any], None, None]:
-        if self.query.from_ == "issues":
-            iterator = self._get_issues
-        else:
+        sources = get_installed_sources()
+
+        try:
+            iterator = sources[self.query.from_]
+        except KeyError:
             raise NotImplementedError(
                 f"No search for source {self.query.from_} implemented."
             )
@@ -532,13 +538,8 @@ class Executor:
         phases: List[
             Tuple[
                 Callable[
-                    [
-                        Generator[Result, None, None],
-                        TaskID,
-                        CounterChannel,
-                        CounterChannel,
-                    ],
-                    Generator[Result, None, None],
+                    [Iterator[Result], TaskID, CounterChannel, CounterChannel,],
+                    Iterator[Result],
                 ],
                 TaskID,
             ]
@@ -560,7 +561,7 @@ class Executor:
         # Link up each generator with its source; these will vary
         # depending on what query feature are in use
         channel = CounterChannel()
-        cursor: Generator = iterator(iterator_task, channel)
+        cursor: Iterator = self._get_cached(iterator(self, iterator_task, channel))
         for phase, task_id in phases:
             output_channel = CounterChannel()
             cursor = phase(cursor, task_id, channel, output_channel)
@@ -575,6 +576,14 @@ class Executor:
     def progress(self) -> Union[Progress, NullProgressbar]:
         return self._progress_bar
 
+    def evaluate_expression(self, row: Result, expression: Expression) -> Any:
+        return row.evaluate_expression(
+            expression,
+            self.query.group_by,
+            functions=self.functions,
+            field_name_map=self.field_name_map,
+        )
+
     def __iter__(self) -> Generator[Dict[str, Any], None, None]:
         progress_bar_cls: Union[Type[Progress], Type[NullProgressbar]] = NullProgressbar
         if self._progress_bar_enabled:
